@@ -7,8 +7,9 @@ import time
 import pytest
 from itertools import chain
 from s3fs.core import S3FileSystem
-from s3fs.utils import seek_delimiter, ignoring, tmpfile, SSEParams
+from s3fs.utils import seek_delimiter, ignoring, SSEParams
 import moto
+import boto3
 
 from botocore.exceptions import NoCredentialsError
 
@@ -250,9 +251,9 @@ def test_ls_touch(s3):
     s3.touch(a)
     s3.touch(b)
     L = s3.ls(test_bucket_name+'/tmp/test', True)
-    assert set(d['Key'] for d in L) == set([a, b])
+    assert {d['Key'] for d in L} == {a, b}
     L = s3.ls(test_bucket_name+'/tmp/test', False)
-    assert set(L) == set([a, b])
+    assert set(L) == {a, b}
 
 
 def test_rm(s3):
@@ -294,6 +295,31 @@ def test_rmdir(s3):
     assert bucket not in s3.ls('/')
 
 
+def test_mkdir(s3):
+    bucket = 'test1_bucket'
+    s3.mkdir(bucket)
+    assert bucket in s3.ls('/')
+
+
+def test_mkdir_region_name(s3):
+    bucket = 'test1_bucket'
+    s3.mkdir(bucket, region_name="eu-central-1")
+    assert bucket in s3.ls('/')
+
+
+def test_mkdir_client_region_name():
+    bucket = 'test1_bucket'
+    try:
+        m = moto.mock_s3()
+        m.start()
+        s3 = S3FileSystem(anon=False, client_kwargs={"region_name":
+                                                     "eu-central-1"})
+        s3.mkdir(bucket)
+        assert bucket in s3.ls('/')
+    finally:
+        m.stop()
+
+
 def test_bulk_delete(s3):
     with pytest.raises((OSError, IOError)):
         s3.bulk_delete(['nonexistent/file'])
@@ -410,8 +436,8 @@ def test_read_keys_from_bucket(s3):
         file_contents = s3.cat('/'.join([test_bucket_name, k]))
         assert file_contents == data
 
-    assert (s3.cat('/'.join([test_bucket_name, k])) ==
-            s3.cat('s3://' + '/'.join([test_bucket_name, k])))
+        assert (s3.cat('/'.join([test_bucket_name, k])) ==
+                s3.cat('s3://' + '/'.join([test_bucket_name, k])))
 
 
 def test_url(s3):
@@ -472,14 +498,15 @@ def test_move(s3):
     assert not s3.exists(fn)
 
 
-def test_get_put(s3):
-    with tmpfile() as fn:
-        s3.get(test_bucket_name+'/test/accounts.1.json', fn)
-        data = files['test/accounts.1.json']
-        assert open(fn, 'rb').read() == data
-        s3.put(fn, test_bucket_name+'/temp')
-        assert s3.du(test_bucket_name+'/temp')[test_bucket_name+'/temp'] == len(data)
-        assert s3.cat(test_bucket_name+'/temp') == data
+def test_get_put(s3, tmpdir):
+    test_file = str(tmpdir.join('test.json'))
+
+    s3.get(test_bucket_name+'/test/accounts.1.json', test_file)
+    data = files['test/accounts.1.json']
+    assert open(test_file, 'rb').read() == data
+    s3.put(test_file, test_bucket_name+'/temp')
+    assert s3.du(test_bucket_name+'/temp')[test_bucket_name+'/temp'] == len(data)
+    assert s3.cat(test_bucket_name+'/temp') == data
 
 
 def test_errors(s3):
@@ -615,6 +642,7 @@ def test_write_small_secure(s3):
     assert s3.cat(secure_bucket_name+'/test') == b'hello'
     head = s3.s3.head_object(Bucket=secure_bucket_name, Key='test')
 
+
 def test_write_large_secure(s3):
     mock = moto.mock_s3()
     mock.start()
@@ -630,8 +658,6 @@ def test_write_large_secure(s3):
 
 
 def test_write_fails(s3):
-    with pytest.raises(NotImplementedError):
-        s3.open(test_bucket_name+'/temp', 'w')
     with pytest.raises(ValueError):
         s3.touch(test_bucket_name+'/temp')
         s3.open(test_bucket_name+'/temp', 'rb').write(b'hello')
@@ -737,6 +763,7 @@ def test_readline_blocksize(s3):
 
         result = f.readline()
         expected = b'ab'
+        assert result == expected
 
 
 def test_next(s3):
@@ -1015,3 +1042,111 @@ def test_versions_unaware(s3):
     with pytest.raises(ValueError):
         with s3.open(versioned_file, version_id='0'):
             fo.read()
+
+
+def test_text_io__stream_wrapper_works(s3):
+    """Ensure using TextIOWrapper works."""
+    s3.mkdir('bucket')
+
+    with s3.open('bucket/file.txt', 'wb') as fd:
+        fd.write(u'\u00af\\_(\u30c4)_/\u00af'.encode('utf-16-le'))
+
+    with s3.open('bucket/file.txt', 'rb') as fd:
+        with io.TextIOWrapper(fd, 'utf-16-le') as stream:
+            assert stream.readline() == u'\u00af\\_(\u30c4)_/\u00af'
+
+
+def test_text_io__basic(s3):
+    """Text mode is now allowed."""
+    s3.mkdir('bucket')
+
+    with s3.open('bucket/file.txt', 'w') as fd:
+        fd.write(u'\u00af\\_(\u30c4)_/\u00af')
+
+    with s3.open('bucket/file.txt', 'r') as fd:
+        assert fd.read() == u'\u00af\\_(\u30c4)_/\u00af'
+
+
+def test_text_io__override_encoding(s3):
+    """Allow overriding the default text encoding."""
+    s3.mkdir('bucket')
+
+    with s3.open('bucket/file.txt', 'w', encoding='ibm500') as fd:
+        fd.write(u'Hello, World!')
+
+    with s3.open('bucket/file.txt', 'r', encoding='ibm500') as fd:
+        assert fd.read() == u'Hello, World!'
+
+
+def test_readinto(s3):
+    s3.mkdir('bucket')
+
+    with s3.open('bucket/file.txt', 'wb') as fd:
+        fd.write(b'Hello, World!')
+
+    contents = bytearray()
+
+    with s3.open('bucket/file.txt', 'rb') as fd:
+        assert fd.readinto(contents) == 13
+
+    assert contents == b'Hello, World!'
+
+
+def test_change_defaults_only_subsequent(monkeypatch):
+    """Test for Issue #135
+
+    Ensure that changing the default block size doesn't affect existing file
+    systems that were created using that default. It should only affect file
+    systems created after the change.
+    """
+    fs_default = S3FileSystem()
+    assert fs_default.default_block_size == 5 * (1024 ** 2)
+
+    fs_overridden = S3FileSystem(default_block_size=64 * (1024 ** 2))
+    assert fs_overridden.default_block_size == 64 * (1024 ** 2)
+
+    # Suppose I want all subsequent file systems to have a block size of 1 GiB
+    # instead of 5 MiB:
+    monkeypatch.setattr(S3FileSystem, 'default_block_size', 1024 ** 3)
+
+    fs_big = S3FileSystem()
+    assert fs_big.default_block_size == 1024 ** 3
+
+    # Test the other file systems created to see if their block sizes changed
+    assert fs_overridden.default_block_size == 64 * (1024 ** 2)
+    assert fs_default.default_block_size == 5 * (1024 ** 2)
+
+
+
+def test_passed_in_session_set_correctly(s3):
+    session = boto3.session.Session()
+    s3 = S3FileSystem(session=session)
+    assert s3.passed_in_session is session
+    client = s3.connect()
+    assert s3.session is session
+
+
+def test_without_passed_in_session_set_unique(s3):
+    session = boto3.session.Session()
+    s3 = S3FileSystem()
+    assert s3.passed_in_session is None
+    client = s3.connect()
+    assert s3.session is not session
+
+
+def test_pickle_without_passed_in_session(s3):
+    s3 = S3FileSystem()
+    try:
+        s3.__getstate__()
+    except NotImplementedError:
+        pytest.fail("Unexpected NotImplementedError")
+
+
+def test_pickle_with_passed_in_session(s3):
+    session = boto3.session.Session()
+    s3 = S3FileSystem(session=session)
+    with pytest.raises(NotImplementedError):
+        s3.__getstate__()
+
+
+
